/*
 * Copyright 2004, Red Hat, Inc. All Rights Reserved
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * on the rights to use, copy, modify, merge, publish, distribute, sub
 * license, and/or sell copies of the Software, and to permit persons to whom
 * the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
 * RED HAT OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Author:
 *    Owen Taylor <otaylor@redhat.com>
 */
#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "texprof-lib.h"
#include "utils.h"

#define INITIAL_ARRAY_SIZE 16

struct Profile_s
{
    enum {
	STATE_NEW,
	STATE_HELLO_WAITING,	/* Waiting for HELLO response */
	STATE_HELLO_RESPONSE,	/* Inside HELLO response */
	STATE_HELLO_HEAP,	/* Inside HEAP inside HELLO */
	STATE_HELLO_TEXTURE,	/* Inside TEXTURE inside HELLO */
	STATE_IDLE,	
	STATE_UPDATE_RESPONSE,	/* Waiting for UPDATE response */
	STATE_UPDATE_WAITING,	/* Inside UPDATE response */
	STATE_UPDATE_HEAP,	/* Inside HEAP inside UPDATE */
	STATE_UPDATE_TEXTURE,	/* Inside TEXTURE inside UPDATE */
    } state;

    char *profilerName;		/* Name of profiler */
    char *targetName;		/* Name received from target application */
  
    ProfileFuncs funcs;		/* Callback data */
    void *userData;

    /* Current list of heaps */
    ProfileHeap **heaps;
    int n_heaps;
    int heaps_space;
    
    ProfileHeap *currentHeap;   /* Heap being updated */
    xbool currentHeapNew;	/* If currentHeap is newly created */
  
    /* Current list of textures */
    ProfileTexture **textures;
    int n_textures;
    int textures_space;

    ProfileTexture *currentTexture; /* Texture being updated */
    xbool currentTextureNew;	    /* if currentTexture is newly created */

    struct timeval updateTimestamp;
};

/* Utility function used when growing arrays by power-of-two */
static int
findBiggerSize (int current,
		int desired)
{
    unsigned int result;
    
    if (current == 0)
	result = INITIAL_ARRAY_SIZE;
    else
	result = current;
    
    while (result < desired && result < INT_MAX)
	result *= 2;
    
    if (result > INT_MAX)
	fatal("Out of memory");

    return (int)result;
}

/************************************************************************
 * Heap management
 ************************************************************************/

static void
growHeaps (Profile *profile,
	   int      new_size)
{
    if (new_size > profile->heaps_space) {
	int new_space = findBiggerSize(profile->n_heaps, new_size);
	profile->heaps = xrealloc(profile->heaps, new_space * sizeof(ProfileHeap *));
	profile->heaps_space = new_space;
    }
}

static ProfileHeap *
newHeap (ProfileID *id)
{
    ProfileHeap *heap = xcalloc(1, sizeof(ProfileHeap));
    heap->id = *id;

    return heap;
}

int
compareIDs (ProfileID *a,
	    ProfileID *b)
{
    if (a->high < b->high)
	return -1;
    else if (a->high == b->high) {
	if (a->low < b->low)
	    return -1;
	else if (a->low == b->low)
	    return 0;
	else
	    return 1;
    } else
	return 1;
}

int
compareIDPointers (const void *aptr,
		   const void *bptr)
{
    ProfileID * const *a = aptr;
    ProfileID * const *b = bptr;

    return compareIDs(*a, *b);
}

static void
addHeap (Profile     *profile,
	 ProfileHeap *heap)
{
    int i;
    int new_indx;

    /* Just to a linear search, since we'd have to write a custom bsearch */
    
    growHeaps(profile, profile->n_heaps + 1);

    for (i = profile->n_heaps - 1; i >= 0; i--)
    if (compareIDs(&profile->heaps[i]->id, &heap->id) < 0)
	break;

    new_indx = i + 1;

    for (i = profile->n_heaps; i > new_indx; i--)
	profile->heaps[i] = profile->heaps[i - 1];

    profile->heaps[new_indx] = heap;
    profile->n_heaps++;

    if (profile->funcs.heapAdded)
	profile->funcs.heapAdded(profile, heap);

}

static void
updateHeap (Profile     *profile,
	    ProfileHeap *heap)
{
    if (profile->funcs.heapUpdated)
	profile->funcs.heapUpdated(profile, heap);
}

static int
lookupHeapIndex (Profile   *profile,
		 ProfileID *id)
{
    ProfileHeap **found;

    if (!profile->n_heaps)
	return -1;

    found = bsearch(&id, profile->heaps, 
		    profile->n_heaps, sizeof(ProfileHeap *),
		    compareIDPointers);
    if (!found)
	return -1;
    else
	return found - profile->heaps;
}

static ProfileHeap *
lookupHeap (Profile   *profile,
	    ProfileID *id)
{
    int indx = lookupHeapIndex(profile, id);
    if (indx == -1)
	return NULL;
    else
	return profile->heaps[indx];
}

static void
destroyHeap (ProfileHeap *heap)
{
    if (heap->name)
	free(heap->name);
    free(heap);
}

static void
deleteHeap (Profile     *profile,
	    ProfileHeap *heap)
{
    int indx = lookupHeapIndex(profile, &heap->id);
    int i;

    assert (indx >= 0);

    for (i = indx; i < profile->n_heaps; i++)
	profile->heaps[i] = profile->heaps[i + 1];

    profile->n_heaps--;

    if (profile->funcs.heapDeleted)
	profile->funcs.heapDeleted(profile, heap);
    
    destroyHeap(heap);
}

/************************************************************************
 * Texture management
 ************************************************************************/

static void
growTextures (Profile *profile,
	      int      new_size)
{
    if (new_size > profile->textures_space) {
	int new_space = findBiggerSize(profile->n_textures, new_size);
	profile->textures = xrealloc(profile->textures, new_space * sizeof(ProfileTexture *));
	profile->textures_space = new_space;
    }
}

static ProfileTexture *
newTexture (ProfileID *id)
{
    ProfileTexture *texture = xcalloc(1, sizeof(ProfileTexture));
    texture->id = *id;

    return texture;
}

static void
addTexture (Profile        *profile,
	    ProfileTexture *texture)
{
    int i;
    int new_indx;

    /* Just use a linear search, since we'd have to write a custom bsearch
     * to find the insertion position for a new element, and inserting
     * is linear anyways
     */
    growTextures(profile, profile->n_textures + 1);

    for (i = profile->n_textures - 1; i >= 0; i--)
	if (compareIDs(&profile->textures[i]->id, &texture->id) < 0)
	    break;

    new_indx = i + 1;

    for (i = profile->n_textures; i > new_indx; i--)
	profile->textures[i] = profile->textures[i - 1];

    profile->textures[new_indx] = texture;
    profile->n_textures++;
    
    if (profile->funcs.textureAdded)
	profile->funcs.textureAdded(profile, texture);
}

static void
updateTexture (Profile        *profile,
	       ProfileTexture *texture)
{
    if (profile->funcs.textureUpdated)
	profile->funcs.textureUpdated(profile, texture);
}

static int
lookupTextureIndex (Profile   *profile,
		    ProfileID *id)
{
    ProfileTexture **found;

    if (!profile->n_textures)
	return -1;

    found = bsearch(&id, profile->textures,
		    profile->n_textures, sizeof(ProfileTexture *),
		    compareIDPointers);
    if (!found)
	return -1;
    else
	return found - profile->textures;
}

static ProfileTexture *
lookupTexture (Profile   *profile,
	       ProfileID *id)
{
    int indx = lookupTextureIndex(profile, id);
    if (indx == -1)
	return NULL;
    else
	return profile->textures[indx];
}

static void
destroyTexture (ProfileTexture *texture)
{
    free(texture);
}

static void
deleteTexture (Profile        *profile,
	       ProfileTexture *texture)
{
    int indx = lookupTextureIndex(profile, &texture->id);
    int i;

    assert(indx >= 0);

    for (i = indx; i < profile->n_textures; i++)
	profile->textures[i] = profile->textures[i + 1];

    profile->n_textures--;

    if (profile->funcs.textureDeleted)
	profile->funcs.textureDeleted(profile, texture);
    
    destroyTexture(texture);
}

/* Checks to see if the heap is non-empty */
static xbool
haveLiveTextures (Profile     *profile,
		  ProfileHeap *heap)
{
    int i;

    for (i = 0; i < profile->n_textures; i++) {
	if (profile->textures[i]->heap == heap)
	    return TRUE;
    }

    return FALSE;
}

/************************************************************************/

/* return TRUE if 'line' starts with 'token' */
static xbool
lineStarts (const char *line,
	    const char *token)
{
    size_t i;

    for (i = 0; token[i]; i++)
	if (line[i] != token[i])
	    return FALSE;

    if (line[i] != '\0' && line[i] != ' ' && line[i] != '=')
	return FALSE;

    return TRUE;
}

/* advance str over digits, return TRUE if there was at least one */
static xbool
skip_digits (const char **str)
{
    const char *start = *str;
    const char *p = *str;

    while ((*p >= '0' && *p <= '9'))
	p++;

    *str = p;
    return p != start;
}

/* advance str over hex digits, return TRUE if there was at least one */
static xbool
skip_hex (const char **str)
{
    const char *start = *str;
    const char *p = *str;

    while ((*p >= '0' && *p <= '9') ||
	   (*p >= 'a' && *p <= 'f') ||
	   (*p >= 'a' && *p <= 'f'))
	p++;
    
    *str = p;
    return p != start;
}

/* advance str over whitespace, return TRUE if there was at least one */
static xbool
skip_spaces (const char **str)
{
    const char *start = *str;
    const char *p = *str;

    while (*p == ' ')
	p++;

    *str = p;
    return p != start;
}

/* Read hex number, up to a certain number of digits */
static unsigned
parseHex (const char *str,
	  int         max_digits)
{
    unsigned result = 0;
    const char *p = str;
    
    while (p < str + max_digits) {
	result *= 16;
	if (*p >= '0' && *p <= '9')
	    result += *p - '0';
	else if (*p >= 'A' && *p <= 'F')
	    result += 10 + *p - 'A';
	else if (*p >= 'a' && *p <= 'f')
	    result += 10 + *p - 'a';
	else
	    break;

	p++;
    }

    return result;
}

/* scanf() style parser, supporting
 *  %i: 64-bit ID
 *  %s: string
 *  %t: timestamp
 *  %u: unsigned integer
 *
 * as defined by the profiling protocol. ' ' represents one or more
 * space. '=' can be surrounded by spaces. Everything else must
 * be matched literally. Trailing whitespace is permitted. */
static xbool
parseLine (const char *line,
	   const char *format,
	   ...)
{
    const char *l;
    const char *f;
    xbool need_alloc = FALSE;
    xbool do_alloc = FALSE;
    va_list vap;

 restart:
    f = format;
    l = line;
    va_start (vap, format);
    for (f = format; *f; f++) {
	switch (*f) {
	case ' ':		/* one or more space */
	    if (!skip_spaces (&l))
		goto error;
	    break;
	case '=':		/* equal sign, optional surrounding spaces */
	    skip_spaces(&l);
	    if (*(l++) != '=')
		goto error;
	    skip_spaces(&l);
	    break; 
	case '%':		/* field to parse */
	    f++;
	    switch (*f) {
	    case 'i': {		/* 64-bit ID */
		    ProfileID *ip = va_arg(vap, ProfileID *);
		    const char *start = l;
		    if (!skip_hex(&l))
			goto error;
		    if (l - start > 16)
			goto error;
		    if (*l && *l != ' ')
			goto error;

		    /* Can't use scanf here because of the way we want to break
		     * up the 16-digit number into two 8-digit numbers in a dynamic way
		     */
		    if (l - start > 8) {
			ip->high = parseHex(start, (l - 8) - start);
			ip->low = parseHex(l - 8, 8);
		    } else {
			ip->high = 0;
			ip->low = parseHex(start, 8);
		    }
		    break;
		}
	    case 's': {
		    char **sp = va_arg(vap, char **);
		    const char *start = l;
		    while (*l) {
			if (*l < ' ' || *l > '~') /* Assume ASCII ordering */
			    goto error;
			l++;
		    }
		    while (l > start && (*(l - 1) == ' '))
			l--;
		    if (l == start)
			goto error;
		    
		    if (do_alloc) {
			if (*sp)
			    free(*sp);
			*sp = xmalloc(l - start + 1);
			memcpy(*sp, start, l - start);
			(*sp)[l - start] = '\0';
		    } else {
			need_alloc = TRUE;
		    }
		    break;
		}
	    case 't': {
		    struct timeval *tv = va_arg(vap, struct timeval *);
		    const char *start = l;
		    unsigned sec, usec;
		    
		    if (!skip_digits(&l))
			goto error;
		    if (*(l++) != '.')
			goto error;
		    if (!skip_digits(&l))
			goto error;
		    if (*l && *l != ' ')
			goto error;
		    
		    sscanf(start, "%u.%u", &sec, &usec);
		    tv->tv_sec = sec;
		    tv->tv_usec = usec;
		    break;
		}
	    case 'u': {
		    unsigned *up = va_arg(vap, unsigned *);
		    const char *start = l;
		    if (!skip_digits(&l))
			goto error;
		    if (*l && *l != ' ')
			goto error;
		    sscanf(start, "%u", up);
		    break;
	        }
	    case '%': {
		    if (*(l++) != '%')
			goto error;
		    break;
		}
	    case '\0': {
		    fprintf(stderr, "parseLine: Missing format character");
		    goto error;
		}
	    default: {
		    fprintf(stderr, "parseLine unknown format character '%c'", *f);
		    goto error;
		}
	    }
	    break; 
	default:
	    if (*(l++) != *f)
		goto error;
	    
	}
    }
    
    skip_spaces(&l);
    if (*l)
	goto error;
    
    if (need_alloc && !do_alloc) {
	do_alloc = TRUE;
	goto restart;
    }

    return TRUE;

error:
    va_end(vap);
    return FALSE;

}

/* Output a single line with printf() style formatting */
static void
outputLine (Profile    *profile,
	    const char *format,
	    ...)
{
    char buffer[PROFILE_MAX_LEN + 1];
    va_list vap;
    int len;
    
    va_start(vap, format);
    len = vsnprintf(buffer, PROFILE_MAX_LEN + 1, format, vap);
    assert(len < PROFILE_MAX_LEN + 1);
    va_end(vap);

    profile->funcs.outputLine(profile, buffer);
}

/**
 * Processes a single line of input from the application
 *
 * \param profile The profile
 * \param line    Buffer holding the incoming line, terminated with \0.
 *                The caller must strip the \r\n at the end of the line
 * \return        NULL if the call succeeded, or an error message
 *                if the line couldn't be succesfully processed.
 */
const char *
profileFeedInputLine (Profile    *profile,
		      const char *line)
{
    if (parseLine(line, ""))	/* ignore empty lines */
	return NULL;
    
    switch (profile->state) {
    case STATE_NEW:
	return "Unexpected input before start of protocol";
    case STATE_HELLO_WAITING:
	if (lineStarts(line, "HELLO")) {
	    unsigned version;
	    
	    if (!parseLine(line, "HELLO %u %t %s", &version, &profile->updateTimestamp, &profile->targetName))
		return "Cannot parse HELLO response";
	    if (version != 0)
		return "Bad version in HELLO response";

	    profile->state = STATE_HELLO_RESPONSE;
	    if (profile->funcs.beginUpdate)
		profile->funcs.beginUpdate(profile);
	} else {
	    return "Unexpected response to HELLO";
	}
	break;
    case STATE_HELLO_RESPONSE:
    case STATE_UPDATE_RESPONSE:
	if (lineStarts(line, "HEAP")) {
	    ProfileID id;
	    
	    if (!parseLine(line, "HEAP %i", &id))
		return "Cannot parse HEAP response";

	    profile->currentHeap = lookupHeap(profile, &id);
	    if (!profile->currentHeap) {
		profile->currentHeap = newHeap(&id);
		profile->currentHeapNew = TRUE;
	    }

	    profile->state = (profile->state == STATE_HELLO_RESPONSE ?
			      STATE_HELLO_HEAP : STATE_UPDATE_HEAP);
	    
	} else if (lineStarts(line, "TEXTURE")) {
	    ProfileID id;
	    
	    if (!parseLine(line, "TEXTURE %i", &id))
		return "Cannot parse TEXTURE response";
	    
	    profile->currentTexture = lookupTexture(profile, &id);
	    if (!profile->currentTexture) {
		profile->currentTexture = newTexture(&id);
		profile->currentTextureNew = TRUE;
	    }
	    
	    profile->state = (profile->state == STATE_HELLO_RESPONSE ?
			      STATE_HELLO_TEXTURE : STATE_UPDATE_TEXTURE);
	    
	} else if (lineStarts(line, "DELETE")) {
	    ProfileHeap *heap;
	    ProfileTexture *texture;
	    ProfileID id;

	    if (!parseLine(line, "DELETE %i", &id))
		return "Cannot parse DELETE response";

	    if ((heap = lookupHeap(profile, &id))) {
		if (haveLiveTextures(profile, heap))
		    return "DELETE of heap with live textures";
		deleteHeap(profile, heap);
	    } else if ((texture = lookupTexture(profile, &id))) {
		deleteTexture(profile, texture);
	    } else {
		return "Unknown ID in DELETE response";
	    }

	} else if (lineStarts(line, "END")) {
	    if (profile->state == STATE_HELLO_RESPONSE) {
		if (!parseLine(line, "END HELLO"))
		    return "HELLO response does not end with END HELLO";
	    } else {
		if (!parseLine(line, "END UPDATE"))
		    return "UPDATE response does not end with END UPDATE";
	    }

	    profile->state = STATE_IDLE;
	    if (profile->funcs.endUpdate)
		profile->funcs.endUpdate(profile);

	} else {
	    if (profile->state == STATE_HELLO_RESPONSE)
		return "Unexpected content in HELLO";
	    else
		return "Unexpected content in UPDATE";
	}
	break;
    case STATE_HELLO_HEAP:
    case STATE_UPDATE_HEAP:
	assert(profile->currentHeap != NULL);
	if (lineStarts(line, "NAME")) {
	    if (!parseLine(line, "NAME=%s", &profile->currentHeap->name))
		return "Cannot parse NAME=";
	    profile->currentHeap->flags |= PROFILE_HEAP_NAME;
	} else if (lineStarts(line, "SIZE")) {
	    if (!parseLine(line, "SIZE=%u", &profile->currentHeap->size))
		return "Cannot parse NAME=";
	    profile->currentHeap->flags |= PROFILE_HEAP_SIZE;
	} else if (lineStarts(line, "END")) {
	    if (!parseLine(line, "END HEAP"))
		return "HEAP does not end with END HEAP";
	    
	    if (profile->currentHeapNew) {
		addHeap(profile, profile->currentHeap);
		profile->currentHeap = NULL;
		profile->currentHeapNew = FALSE;
	    } else {
		updateHeap(profile, profile->currentHeap);
		profile->currentHeap = NULL;
	    }
	    
	    profile->state = (profile->state == STATE_HELLO_HEAP ?
			      STATE_HELLO_RESPONSE : STATE_UPDATE_RESPONSE);
	} else {
	    return "Unexpected content in HEAP";
	}
	break;
    case STATE_HELLO_TEXTURE:
    case STATE_UPDATE_TEXTURE:
	assert(profile->currentTexture != NULL);
	if (lineStarts(line, "WIDTH")) {
	    if (!parseLine(line, "WIDTH=%u", &profile->currentTexture->width))
		return "Cannot parse WIDTH=";
	    profile->currentTexture->flags |= PROFILE_TEXTURE_WIDTH;
	} else if (lineStarts(line, "HEIGHT")) {
	    if (!parseLine(line, "HEIGHT=%u", &profile->currentTexture->height))
		return "Cannot parse HEIGHT=";
	    profile->currentTexture->flags |= PROFILE_TEXTURE_HEIGHT;
	} else if (lineStarts(line, "MAIN_SIZE")) {
	    if (!parseLine(line, "MAIN_SIZE=%u", &profile->currentTexture->mainSize))
		return "Cannot parse MAIN_SIZE=";
	    profile->currentTexture->flags |= PROFILE_TEXTURE_MAIN_SIZE;
	} else if (lineStarts(line, "HEAP")) {
	    ProfileID id;
	    ProfileHeap *heap;

	    if (parseLine(line, "HEAP=", &id)) {
		profile->currentTexture->heap = NULL;
		profile->currentTexture->flags |= PROFILE_TEXTURE_HEAP;
		return NULL;
	    }
		
	    if (!parseLine(line, "HEAP=%i", &id))
		return "Cannot parse HEAP=";
		
	    if ((heap = lookupHeap(profile, &id))) {
		profile->currentTexture->heap = heap;
		profile->currentTexture->flags |= PROFILE_TEXTURE_HEAP;
	    } else {
		return "Unknown ID in HEAP=";
	    }
	    
	} else if (lineStarts(line, "HEAP_TS")) {
	    if (!parseLine(line, "HEAP_TS=%t", &profile->currentTexture->heapTS))
		return "Cannot parse HEAP_TS=";
	    profile->currentTexture->flags |= PROFILE_TEXTURE_HEAP_TS;
	} else if (lineStarts(line, "HEAP_SIZE")) {
	    if (!parseLine(line, "HEAP_SIZE=%u", &profile->currentTexture->heapSize))
		return "Cannot parse HEAP_SIZE=";
	    profile->currentTexture->flags |= PROFILE_TEXTURE_HEAP_SIZE;
	} else if (lineStarts(line, "END")) {
	    if (!parseLine(line, "END TEXTURE"))
		return "TEXTURE does not end with END TEXTURE";

	    if (profile->currentTextureNew) {
		addTexture(profile, profile->currentTexture);
		profile->currentTexture = NULL;
		profile->currentTextureNew = FALSE;
	    } else {
		updateTexture(profile, profile->currentTexture);
		profile->currentTexture = NULL;
	    }
	    
	    profile->state = (profile->state == STATE_HELLO_TEXTURE ?
			      STATE_HELLO_RESPONSE : STATE_UPDATE_RESPONSE);
	} else {
	    return "Unexpected content in TEXTURE";
	}
	break;
    case STATE_IDLE:
	return "Unexpected input while not waiting for a response";
    case STATE_UPDATE_WAITING:
	if (lineStarts(line, "UPDATE")) {
	    if (!parseLine(line, "UPDATE %t", &profile->updateTimestamp))
		return "Cannot parse HELLO response";
	} else {
	    return "Unexpected response to UPDATE";
	}
	profile->state = STATE_UPDATE_RESPONSE;
	if (profile->funcs.beginUpdate)
	    profile->funcs.beginUpdate(profile);
	break;
    }

    return NULL;
}
    
/**
 * Creates a new profile object. The caller provides a structure
 * of callback functions. The size of the structure is provided
 * to allow for compatible extension of the set of callback functions.
 *
 * \param profilerName  The profile
 * \param funcs         Callback functions
 * \param funcs_size    size of funcs in bytes
 * \param user_data     user data
 * \return              the new profile
 *
 * \sa profileDestroy
 */
Profile *
profileNew (const char         *profilerName,
	    const ProfileFuncs *funcs,
	    size_t              funcs_size,
	    void               *user_data)
{
    Profile *profile = xcalloc(1, sizeof(Profile));

    profile->profilerName = xstrdup(profilerName);
    
    if (funcs_size > sizeof(ProfileFuncs))
	funcs_size = sizeof(ProfileFuncs);
    
    memcpy((char *)&profile->funcs, (char *)funcs, funcs_size);
    assert(profile->funcs.outputLine);

    return profile;
}

/**
 * Returns the user data provided to profileNew.
 *
 * \param profile  The profile
 * \return         the user data provided to profileNew
 */
void *
profileUserData (Profile *profile)
{
    return profile->userData;
}

/**
 * Requests an update of the profile. When the update finishes,
 * the endUpdate callback function will be called.
 *
 * \param profile  The profile
 * \return         TRUE if an update will be received, FALSE
 *                 if the profile was still in the process of processing
 *                 the previous request.
 */
int
profileUpdate (Profile *profile)
{
    switch (profile->state) {
    case STATE_NEW:
	outputLine (profile, "HELLO 0 %s",
		    profile->profilerName ? profile->profilerName : "texprof-lib");
	profile->state = STATE_HELLO_WAITING;
	return TRUE;
    case STATE_IDLE:
	outputLine (profile, "UPDATE");
	profile->state = STATE_UPDATE_WAITING;
	return TRUE;
    default:
	return FALSE;
    }
}

/**
 * Destroys the profile and all associated objects
 *
 * \param profile  The profile
 */
void
profileDestroy (Profile *profile)
{
    int i;

    for (i = 0; i < profile->n_textures; i++)
	destroyTexture(profile->textures[i]);
    if (profile->currentHeap && profile->currentHeapNew)
	destroyHeap(profile->currentHeap);
    
    for (i = 0; i < profile->n_heaps; i++)
	destroyHeap(profile->heaps[i]);
    if (profile->currentHeap && profile->currentHeapNew)
	destroyHeap(profile->currentHeap);

    if (profile->profilerName)
	free(profile->profilerName);
    
    free(profile);
}

/**
 * Gets the application timestamp for the last time an update
 * of the profile was received from the application. This is
 * the appropriate value to compare to the timestamps of
 * individual objects in the profile.
 *
 * \param profile  The profile
 * \param tv       Location to store the timestamp
 */
void
profileGetUpdateTimestamp (Profile        *profile,
			   struct timeval *tv)
{
    if (tv)
	*tv = profile->updateTimestamp;
}

/**
 * Lists the current heaps for the profile. The resulting
 * array is owned by the profile and must not be modified or freed.
 *
 * \param profile  The profile
 * \param heaps    Location to store the array of heaps
 * \param n_heaps  Location to store number of heaps
 */
void
profileGetHeaps (Profile               *profile,
		 ProfileHeap * const  **heaps,
		 int                   *n_heaps)
{
    if (n_heaps)
	*n_heaps = profile->n_heaps;
    if (heaps)
	*heaps = profile->heaps;
}

/**
 * Lists the current textures for the profile. The resulting
 * array is owned by the profile and must not be modified or freed.
 *
 * \param profile    The profile
 * \param textures   Location to store the array of textures
 * \param n_textures Location to store number of textures
 */
void
profileGetTextures (Profile                  *profile,
		    ProfileTexture * const **textures,
		    int                      *n_textures)
{
    if (n_textures)
	*n_textures = profile->n_textures;
    if (textures)
	*textures = profile->textures;
}
