/*
 * Copyright 2004, Red Hat, Inc. All Rights Reserved
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * on the rights to use, copy, modify, merge, publish, distribute, sub
 * license, and/or sell copies of the Software, and to permit persons to whom
 * the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
 * RED HAT OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Author:
 *    Owen Taylor <otaylor@redhat.com>
 */
#ifndef TEXPROF_LIB_H
#define TEXPROF_LIB_H

#include <sys/time.h>

typedef struct Profile_s Profile;
typedef struct ProfileFuncs_s ProfileFuncs;
typedef struct ProfileHeap_s ProfileHeap;
typedef struct ProfileID_s ProfileID;
typedef struct ProfileTexture_s ProfileTexture;

/* From the protocol */
#define PROFILE_MAX_LEN 1024

/* 64-bit ID, avoid using C99 uint64_t */
struct ProfileID_s
{
    unsigned high;
    unsigned low;
};

/* Flags for what fields are set in ProfileHeap structure */
typedef enum
{
  PROFILE_HEAP_NAME = 1 << 0,
  PROFILE_HEAP_SIZE = 1 << 1
} ProfileHeapFlags;

struct ProfileHeap_s
{
    ProfileID id;
    ProfileHeapFlags flags;

    char *name;
    unsigned size;
};

/* Flags for what fields are set in ProfileTexture structure */
typedef enum
{
  PROFILE_TEXTURE_WIDTH       = 1 << 0,
  PROFILE_TEXTURE_HEIGHT      = 1 << 1,
  PROFILE_TEXTURE_MAIN_SIZE   = 1 << 2,
  PROFILE_TEXTURE_HEAP        = 1 << 4,
  PROFILE_TEXTURE_HEAP_SIZE   = 1 << 5,
  PROFILE_TEXTURE_HEAP_TS     = 1 << 6
} ProfileTextureFlags;

struct ProfileTexture_s
{
    ProfileID id;
    ProfileTextureFlags flags;
    
    unsigned width;
    unsigned height;
    
    unsigned mainSize;
    
    ProfileHeap *heap;
    struct timeval heapTS;
    unsigned heapSize;
};

struct ProfileFuncs_s
{
    /* Write the line to the output stream. line will be at most
     * PROFILE_MAX_LEN - 2 bytes long, not counting the terminating \0.
     * It does not include the \r\n line separator.
     */
    void (*outputLine)  (Profile   *profile,
			 const char *line);

    /* Called when a batch of updates starts, all calls to
     * {heap,texture}{Added,Updated,Deleted} will occur within
     * beginUpdate/endUpdate pairs. 
     */
    void (*beginUpdate) (Profile *profile);

    /* Called when a batch of updates ends, Each successful call to
     * profileUpdate will cause exactly one call to endUpdate
     */
    void (*endUpdate)   (Profile *profile);

    /* Called when a new heap is added */
    void (*heapAdded) (Profile     *profile,
		       ProfileHeap *heap);
    /* Called when a heap is changed */
    void (*heapUpdated) (Profile     *profile,
			 ProfileHeap *heap);
    /* Called when a heap is deleted */
    void (*heapDeleted) (Profile     *profile,
			 ProfileHeap *heap);
    
    /* Called when a texture is added */
    void (*textureAdded)   (Profile        *profile,
			    ProfileTexture *texture);
    /* Called when a texture is updated */
    void (*textureUpdated) (Profile        *profile,
			    ProfileTexture *texture);
    /* Called when a texture is deleted */
    void (*textureDeleted) (Profile        *profile,
			    ProfileTexture *texture);    
};

Profile *profileNew (const char         *profilerName,
		     const ProfileFuncs *funcs,
		     size_t              funcs_size,
		     void               *user_data);

int profileUpdate (Profile *profile);

void *profileUserData      (Profile *profile);
void  profileDestroy       (Profile *profile);

const char *profileFeedInputLine (Profile    *profile,
				  const char *line);

void  profileGetUpdateTimestamp (Profile        *profile,
				 struct timeval *tv);

void profileGetHeaps    (Profile                 *profile,
			 ProfileHeap * const    **heaps,
			 int                     *n_heaps);
void profileGetTextures (Profile                 *profile,
			 ProfileTexture * const **textures,
			 int                     *n_textures);

#endif /* TEXPROF_LIB_H */
